; Calculating Bezier surfaces;
; Implemented in Flat assembler by Maciej Guba,
; http://macgub.co.pl
; Requires SSE3



; I use this equation:
; x(s,t)=(1-s)^3*(x11*a+x12*b+x13*c+x14*d)+
;	 +3*(1-s)^2*s*(x21*a+x22*b+x23*c+x24*d)+
;	 +3*(1-s)*s^2*(x31*a+x32*b+x33*c+x34*d)+
;	 +s^3*(x41*a+x42*b+x43*c+x44*d)
; y(s,t) and z(s,t) similar...
;  a = (1-t)^3
;  b = 3*(1-t)^2*t
;  c = 3*(1-t)*t^2
;  d = t^3


;		      (x11*a+x12*b+x13*c+x14*d)+
;		      (x21*a+x22*b+x23*c+x24*d)+
;		      (x31*a+x32*b+x33*c+x34*d)+
;		      (x41*a+x42*b+x43*c+x44*d)



;		      ; kind of matrix - I called it as tensor product of
;		      ; x11 - x44
if 0
tens_product:
  ; esi -  a, b, c, d factors
  ; ebx -  x11	-  x44
  ;	   y11	-  y44
  ;	   z11	-  z44
  ; edi - product

   movups xmm0,[ebx]
   mulps  xmm0,[esi]
   haddps xmm0,xmm0
   haddps xmm0,xmm0
   movss  [edi],xmm0




end if


calc_bez_factors:
; in:
; xmm0 - factor value - 't'
; out:
; edi - a, b, c, d

;  a = (1-t)^3
;  b = 3*(1-t)^2*t
;  c = 3*(1-t)*t^2
;  d = t^3


   .aB	equ [edi]
   .bB	equ [edi+4]
   .cB	equ [edi+8]
   .dB	equ [edi+12]


   movss  xmm1,[f1]
   subss  xmm1,xmm0
   movss  xmm2,xmm1
   movss  xmm3,xmm1
   movss  xmm4,xmm0
   mulss  xmm2,xmm1
   mulss  xmm1,[f3]
   mulss  xmm3,xmm2
   movss  .aB,xmm3
   mulss  xmm2,[f3]
   mulss  xmm4,xmm4
   mulss  xmm2,xmm0
   movss  .bB,xmm2
   mulss  xmm1,xmm4
   movss  .cB,xmm1
   mulss  xmm4,xmm0 ;.t
   movss  .dB,xmm4

ret

;============================================================================
bezier_surface:
 ; makes bezier surface through tensor product
	; if eax = 'offd'   - calc only one vertex value  - x,y,z
	;	 xmm0 - t
	;	 xmm1 - s
	; edx - patch qality dimensions  as word integer: x shl 16 + y

	push	 ebp
	mov	 ebp,esp
	sub	 esp,90
	and	 ebp,-16

  .tB		 equ [ebp-4]
  .sB		 equ [ebp-8]
  .ctB		 equ dword [ebp-12] ; integer counter t
  .csB		 equ dword [ebp-16] ; int counter s
  .vert_countBB  equ dword [ebp-20]
  .tri_countBB	 equ dword [ebp-24]


  .x1 equ esi		;      .x41-x44
  .x2 equ esi+16	;      .x31-x34
  .x3 equ esi+32	;      .x21-x24
  .x4 equ esi+48	;      .x11-x14

  .y1 equ esi+64	;      .y41-y44
  .y2 equ esi+80	;      .y31-y34
  .y3 equ esi+96	;      .y21-y24
  .y4 equ esi+112	;      .y11-y14

  .z1 equ eax		;      .z41-z44
  .z2 equ eax+16	;      .z31-z34
  .z3 equ eax+32	;      .z21-z24
  .z4 equ eax+48	;      .z11-z14

  .dsB equ   [ebp-28]	  ; delta s
  .dtB equ   [ebp-32]	  ; delta t

  .d  equ    [ebp-36]
  .c  equ    [ebp-40]
  .b  equ    [ebp-44]
  .a  equ    [ebp-48]
  .abcd equ  [ebp-48]

  .d1  equ   [ebp-52]
  .c1  equ   [ebp-56]
  .b1  equ   [ebp-60]
  .a1  equ   [ebp-64]
  .abcd1 equ [ebp-64]
  .one_v equ	dword[ebp-68]

   mov	    .one_v,eax
   movss    .tB,xmm1
   movss    .sB,xmm0
   xchg     ecx,eax
   lea	    eax,[esi+128]
   cmp	    ecx,'offd'
   je	    .ffd1

   movzx    ecx,dx
   mov	    .csB,ecx
 ;  sub      .cs,1
   cvtsi2ss xmm1,ecx
   rcpss    xmm1,xmm1
   movss    .dsB,xmm1
   shr	    edx,16
   mov	    .ctB,edx
;   sub      .ctB,1
   cvtsi2ss xmm0,edx
   rcpss    xmm0,xmm0
   movss    .dtB,xmm0

 ;  push     edi
 ;  push     ebx


   xor	    ecx,ecx
   mov	    .sB,ecx
   mov	    .tB,ecx
   mov	    .vert_countBB,ecx
   mov	    .tri_countBB,ecx
 .ffd1:
   push      ebx
 .looptt:
 ;  movss    .sB,xmm1
;  .ffd1:
   push     ecx
   push     edi
   lea	    edi,.abcd
   movss    xmm0,.tB
   call     calc_bez_factors ;calc_factors
   pop	    edi
   xor	    ecx,ecx
   cmp	    .one_v,'offd'
   je	    .loopss
   mov	    dword .sB,ecx
 .loopss:
   push     ecx
   movss    xmm0,.sB
   push     edi
   lea	    edi,.abcd1

   call     calc_bez_factors
   pop	    edi
   movups   xmm4,.a1
   movaps   xmm5,xmm4
   movaps   xmm6,xmm4
   movaps   xmm7,xmm4

   mulps    xmm4,[.x1]
   mulps    xmm5,[.x2]
   mulps    xmm6,[.x3]
   mulps    xmm7,[.x4]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   mulss    xmm4,.a
   mulss    xmm5,.b
   mulss    xmm6,.c
   mulss    xmm7,.d
   addss    xmm4,xmm5
   addss    xmm6,xmm7
   addss    xmm4,xmm6
   movss    [edi],xmm4


   movups   xmm4,.a1
   movaps   xmm5,xmm4
   movaps   xmm6,xmm4
   movaps   xmm7,xmm4

   mulps    xmm4,[.y1]
   mulps    xmm5,[.y2]
   mulps    xmm6,[.y3]
   mulps    xmm7,[.y4]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   mulss    xmm4,.a
   mulss    xmm5,.b
   mulss    xmm6,.c
   mulss    xmm7,.d
   addss    xmm4,xmm5
   addss    xmm6,xmm7
   addss    xmm4,xmm6
   movss    [edi+4],xmm4

   movups   xmm4,.a1
   movaps   xmm5,xmm4
   movaps   xmm6,xmm4
   movaps   xmm7,xmm4

   mulps    xmm4,[.z1]
   mulps    xmm5,[.z2]
   mulps    xmm6,[.z3]
   mulps    xmm7,[.z4]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   mulss    xmm4,.a
   mulss    xmm5,.b
   mulss    xmm6,.c
   mulss    xmm7,.d
   addss    xmm4,xmm5
   addss    xmm6,xmm7
   addss    xmm4,xmm6
   movss    [edi+8],xmm4

   movss    xmm0,.sB
   addss    xmm0,.dsB
   movss    .sB,xmm0
   add	    edi,12
   inc	    .vert_countBB
   pop	    ecx
   inc	    ecx
   movss    xmm7,.csB
   cmp	    .one_v,'offd'
   je	    .break
   cmp	    ecx,.csB
   jb	   .loopss

   movss    xmm1,.tB
   addss    xmm1,.dtB
   movss    .tB,xmm1

   pop	    ecx
   inc	    ecx
   cmp	    ecx,.ctB
   jb	   .looptt


  ; vertices described one bezier surface calculated
  ; calc triangles list

   cld
   mov	  edx,.ctB
   mov	  esi,ebx
   xor	  ebx,ebx
 .oop1:
   mov	  ecx,.csB
  @@:
   mov	  eax,ebx

   mov	  [esi],eax
   inc	  eax
   mov	  [esi+20],eax
   mov	  [esi+8],eax
   add	  eax,.csB
   mov	  [esi+4],eax
   mov	  [esi+12],eax
   inc	  eax
   mov	  [esi+16],eax
   add	  esi,24
   inc	  ebx
   add	  .tri_countBB,2
   loop   @b
   inc	  ebx
   dec	  edx
   jnz	  .oop1

;   pop      ebx
;   pop      edi

   mov	     ebx,.tri_countBB
   imul      ebx,12

   pop	     ecx

;   add       ecx,ecx

   add	     ebx,ecx

  ; ebx - end of computed triangles list ptr
  ; edi - end of calculated vertices ptr

;   mov      ecx,.vert_countB

   mov	    edx,.tri_countBB
   mov	    ecx,.vert_countBB


   add	    esp,90
   pop	    ebp

ret
.break:
    add     esp,98
    pop     ebp
ret

;==========================================================

draw_bezier_derives1:
  ; esi - source  derives nodes
  ; ebx - source derv indexes (rectangles)
  ; edi - screen buffer
  ; .screen equ [ebp-4]
  ; uses some globals
     push     ebp
     mov      ebp,esp
  ;   sub      esp,16
  ;   .counter equ dword[ebp-4]
     emms
;     mov      .screen,edi
     movzx    eax,[xres_var]
     movd     mm2,eax
     movd     mm1,edi  ;.screen
     mov      eax,0x0000ff00
     movd     mm3,eax

     movd     xmm7,dword[xres_var]
     mov      eax,1
     movd     xmm5,eax
     shufps   xmm5,xmm5,0
     pxor     xmm6,xmm6
     punpcklwd xmm7,xmm6
     movlhps  xmm7,xmm7
     psubd    xmm7,xmm5
;     push     esi
     push     ebx
     mov      ecx,4
  .again1:
     push     ebx
     push     ecx
     mov      ecx,3
  .again:
     push     ecx
     push     ebx
     movzx    eax,word[ebx]  ;.counter
     movzx    edx,word[ebx+2] ; .counter
     imul     edx,12
     add      edx,esi
     imul     eax,12
     add      eax,esi
     movlps   xmm0,[eax]
     movhps   xmm0,[edx]
 ;    movaps   xmm3,xmm0
     cvtps2dq xmm0,xmm0
     movaps   xmm2,xmm0
     movaps   xmm1,xmm0
     pcmpgtd  xmm0,xmm7
     pcmpgtd  xmm1,xmm6
     pxor     xmm0,xmm1
     pmovmskb eax,xmm0
     and      ah,al
     cmp      ah,0xff
     jne      .skip1


     ; eax=x1 ebx=y1 ecx=x2 edx=y2
     sub      esp,16
     movups   [esp],xmm2
     pop      eax ebx ecx edx


     pxor     xmm6,xmm6
     movd     xmm7,[xres_vard]
     punpcklwd xmm7,xmm6
     pushad
 ;    movd     mm0,[Zbuffer_ptr]
 ;    pcmpeqd  xmm0,xmm0
 ;    psrldq   xmm0,2
 ;    movaps   xmm1,xmm0
 ;    pxor     xmm2,xmm2
 ;    pxor     xmm3,xmm3
     call     linea  ;_grd ;a
     popad


   .skip1:
     pop     ebx
     add     ebx,2
     pop     ecx
     loop    .again

     pop     ecx
     pop     ebx
     add     ebx,8
     dec     ecx
     jnz   .again1

     pop     ebx

     mov      ecx,4
  .again2:
     push     ecx
     mov      ecx,3
  .again3:
     push     ecx
     push     ebx
     movzx    eax,word[ebx]
     movzx    edx,word[ebx+8]
     imul     eax,12
     imul     edx,12
     movlps   xmm0,[esi+eax]
     movhps   xmm0,[esi+edx]
     cvtps2dq xmm0,xmm0
     movaps   xmm2,xmm0
     movaps   xmm1,xmm0
     pcmpgtd  xmm0,xmm7
     pcmpgtd  xmm1,xmm6
     pxor     xmm0,xmm1
     pmovmskb eax,xmm0
     and      ah,al
     cmp      ah,0xff
     jne      .skip2

     ; eax=x1 ebx=y1 ecx=x2 edx=y2
     pxor     xmm6,xmm6
     movd     eax,xmm2
     psrldq   xmm2,4
     movd     ebx,xmm2
     psrldq   xmm2,4
     movd     ecx,xmm2
     psrldq   xmm2,4
     movd     edx,xmm2
     movd     xmm7,[xres_vard]
     punpcklwd xmm7,xmm6

     pushad
     call     linea
     popad
   .skip2:
     pop     ebx
     add     ebx,2
     pop     ecx
     dec     ecx
     jnz     .again3
     pop     ecx
     dec     ecx
     jnz     .again2

     mov   esp,ebp
     pop   ebp

ret
;====================================================

draw_bezier_bars:
;ret
; in:
;  esi - ptr to derives vertices list
;  ebx - ptr to derives indexes (16 pro one rectangle)
;  ecx - number of surfaces
;  globals - screen_ptr
   push   ebp
   mov	  ebp,esp
   .ym4   equ dword[ebp-4]

   shl	    ecx,4
   movzx    edx,[xres_var]
   movzx    edi,[yres_var]
   sub	    edx,4
   sub	    edi,4
   push     edi

   pxor     xmm0,xmm0
 .oop:
   push   ecx
   movzx  eax,word[ebx]
   movzx  ecx,word[ebx]
   imul   eax,12
   add	  eax,esi
   movlps xmm0,[eax]
   push   ebx
   push   edx
   push   esi
   mov	  esi,.ym4
   mov	  eax,0x00ffff00

 ;  ecx - edge index; signature
 ;  xmm0 - coordinates
 ;  eax - color
 ;  edx  - xresm4
 ;  esi  - yresm4
   call   bar_edge ; edit.inc
   pop	  esi
   pop	  edx
   pop	  ebx
   add	  ebx,2
   pop	  ecx
   loop   .oop

   mov	  esp,ebp
   pop	  ebp
ret
